type sommet = int 
type graphe = sommet list array



(*** Question 1 ***)
let graphe = [|
  [5];
  [2; 3];
  [4; 5];
  [6; 7];
  [9];
  [10; 11];
  [12];
  [12];
  [4; 10];
  [6];
  [0];
  [7];
  []
|]


(*** Question 1.5 : implémentation d'une file ***)
(* On implémente une file de capacité donnée par un tableau circulaire. *)
type file = {
  tableau: int array;
  mutable debut: int;
  mutable taille: int;
}

let creer capacite = {
  tableau = Array.make capacite 0;
  debut = 0;
  taille = 0
}

let est_vide file =
  file.taille = 0

let est_pleine file =
  let n = Array.length file.tableau in 
  file.taille = n

let enfiler file element =
  let n = Array.length file.tableau in 
  if est_pleine file then failwith "La file est pleine"
  else 
    let fin = (file.debut + file.taille) mod n in
    file.tableau.(fin) <- element;
    file.taille <- file.taille + 1

let defiler file =
  let n = Array.length file.tableau in 
  if est_vide file then failwith "La file est vide"
  else
    let element = file.tableau.(file.debut) in
    file.debut <- (file.debut + 1) mod n;
    file.taille <- file.taille - 1;
    element



(*** Question 2 ***)
type couleur = Blanc | Gris | Noir
let parcours_en_largeur graphe s =
  let n = Array.length graphe in
  let couleur = Array.make n Blanc in
  let file = creer n in  (* il n'y aura jamais plus que n sommets dans la file *)
  let res = ref [] in 

  (* on commence par enfile le sommet initial *)
  enfiler file s;
  couleur.(s) <- Gris;

  while not (est_vide file) do
    let u = defiler file in
    res := u :: !res; 

    (* On parcourt les voisins de u *)
    let rec traiter_voisins voisins =
      match voisins with
      | [] -> ()
      | voisin :: voisins' -> 
          if couleur.(voisin) = Blanc then begin
            couleur.(voisin) <- Gris;
            enfiler file voisin
          end;
          traiter_voisins voisins'
    in

    traiter_voisins graphe.(u);
    couleur.(u) <- Noir
  done;

  List.rev !res (* on renverse la liste pour obtenir l'ordre correct *)



(* Tests *)
let _ = parcours_en_largeur graphe 0 (* [0; 5; 10; 11; 7; 12] *)
let _ = parcours_en_largeur graphe 2 (* [2; 4; 5; 9; 10; 11; 6; 0; 7; 12] *)



(*** Question 3 ***)
let distances graphe s =
  let n = Array.length graphe in
  let couleur = Array.make n Blanc in
  let file = creer n in  
  let dist = Array.make n None in 

  enfiler file s;
  couleur.(s) <- Gris;
  dist.(s) <- Some 0; (* Initialisation de la distance à s à 0 *)

  while not (est_vide file) do
    let u = defiler file in

    let rec traiter_voisins voisins =
      match voisins with
      | [] -> ()
      | voisin :: voisins' -> 
          if couleur.(voisin) = Blanc then begin
            couleur.(voisin) <- Gris;
            enfiler file voisin;
            match dist.(u) with
            | None -> failwith "Impossible : le parent n'a pas de distance"
            | Some d -> dist.(voisin) <- Some (d + 1)  (* Mise à jour de la distance *)
          end;
          traiter_voisins voisins'
    in

    traiter_voisins graphe.(u);
    couleur.(u) <- Noir
  done;

  dist 

  
(* Tests *)
let _ = distances graphe 0
let _ = distances graphe 2 
let _ = distances graphe 8 